/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * chat_interface.c
 *  - Text and Entry boxes for chatting
 *  - Maintains list of available chat interfaces
 *  - Functions to output messages to the correct chat interface based
 *    on recipient
 * -- 
 * $Id: chat_interface.c,v 1.5 2001/06/07 12:37:46 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <string.h>

#include <gtk/gtk.h>

#include <meko.h>

#include "chat_interface.h"
#include "connect_dialog.h"
#include "unicode.h"

/* Forward prototype for command functions */
static void user_motd(ChatInterface *, const gchar *);
static void user_prefs(ChatInterface *, const gchar *);
static void user_password(ChatInterface *, const gchar *);
static void user_status(ChatInterface *, const gchar *);
static void user_away(ChatInterface *, const gchar *);
static void user_busy(ChatInterface *, const gchar *);
static void user_here(ChatInterface *, const gchar *);
static void user_nick(ChatInterface *, const gchar *);
static void user_whom(ChatInterface *, const gchar *);
static void user_who(ChatInterface *, const gchar *);
static void user_whois(ChatInterface *, const gchar *);
static void user_seen(ChatInterface *, const gchar *);
static void user_channel(ChatInterface *, const gchar *);
static void user_say(ChatInterface *, const gchar *);
static void user_me(ChatInterface *, const gchar *);
static void user_msg(ChatInterface *, const gchar *);
static void user_act(ChatInterface *, const gchar *);
static void user_wall(ChatInterface *, const gchar *);
static void user_history(ChatInterface *, const gchar *);
static void user_boot(ChatInterface *, const gchar *);
static void user_clear(ChatInterface *, const gchar *);

/* Forward prototypes */
static void chat_interface_entry_activated(GtkWidget *, ChatInterface *);

/* Font for the interface */
#define CHAT_FONT "-*-lucidatypewriter-medium-r-normal-" \
		  "*-*-120-*-*-*-*-iso8859-1"

/* List of currently open interfaces */
GSList *open_interfaces = 0;

/* Command prefix */
static const gchar command_prefix[] = "/";

/* Command dispatch table */
	static DispatchItem command_table[] = {
#if 0
	/* login, connect, guest, logout, discon */
#endif
	{ "motd", &user_motd },
	{ "prefs", &user_prefs },
	{ "pref", &user_prefs },
	{ "password", &user_password },
	{ "pass", &user_password },
	{ "status", &user_status },
	{ "away", &user_away },
	{ "busy", &user_busy },
	{ "here", &user_here },
	{ "unaway", &user_here },
	{ "unbusy", &user_here },
	{ "nick", &user_nick },
	{ "whom", &user_whom },
	{ "who", &user_who },
	{ "members", &user_who },
	{ "whois", &user_whois },
	{ "seen", &user_seen },
	{ "whowas", &user_seen },
	{ "channel", &user_channel },
	{ "chan", &user_channel },
	{ "say", &user_say },
	{ "me", &user_me },
	{ "msg", &user_msg },
	{ "m", &user_msg },
	{ "action", &user_act },
	{ "act", &user_act },
	{ "wall", &user_wall },
	{ "history", &user_history },
	{ "hist", &user_history },
	{ "boot", &user_boot },
	{ "clear", &user_clear },
#if 0
	{ "quit", &user_quit },
	{ "exit", &user_quit },
	{ "raw", &user_raw },
#endif
	{ NULL, NULL }
};

/* Command dispatch table for unconnected interfaces */
static DispatchItem uc_command_table[] = {
	{ NULL, NULL }
};


/* user_motd
 * Handle the /motd command
 */
static void
user_motd(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);
	send_enaml_buildf(ci->server->sock, "motd", "");
}

/* user_prefs
 * Handle the /prefs command
 */
static void
user_prefs(ChatInterface *ci, const gchar *text)
{
	gchar **arg;

	g_assert(ci);

	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** Usage: %spref"
				         "[ <pref> [<value>]]\n",
				     command_prefix);
		send_enaml_buildf(ci->server->sock, "prefs", "");
		return;
	}

	/* Can only change the tagline for now.  Password is changed using
	 * /pass, notify and icon are used by meko
	 */
	if (!g_strcasecmp(arg[0], "tagline")) {
		gchar *text2;

		text2 = utf8_encode(arg[1] ? arg[1] : "");
		send_enaml_buildf(ci->server->sock, "prefs", "tagline:%s");
		g_free(text2);

	} else {
		chat_interface_sayto(ci, "*** Available preferences: %s\n",
				     "tagline");
	}

	g_strfreev(arg);
}

/* user_password
 * Handle the /password command
 */
static void
user_password(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %spassword <new-pass>\n",
				     command_prefix);
		return;
	}

	send_enaml_buildf(ci->server->sock, "prefs", "password:%s", text);
}

/* user_status
 * Handle the /status command
 */
static void
user_status(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);
	send_enaml_buildf(ci->server->sock, "status", "");
}

/* user_away
 * Handle the /away command
 */
static void
user_away(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (strlen(text)) {
		gchar *text2;

		text2 = utf8_encode(text);
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "away", text2);
		g_free(text2);

	} else if (ci->server->status != SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "away", "");
	} else {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}
}

/* user_busy
 * Handle the /busy command
 */
static void
user_busy(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (strlen(text)) {
		gchar *text2;

		text2 = utf8_encode(text);
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "busy", text2);
		g_free(text2);

	} else if (ci->server->status != SAY2_STATUS_BUSY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "busy", "");
	} else {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}
}

/* user_here
 * Handle the /here command
 */
static void
user_here(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (strlen(text)) {
		gchar *text2;

		text2 = utf8_encode(text);
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", text2);
		g_free(text2);
	} else {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}
}

/* user_nick
 * Handle the /nick command
 */
static void
user_nick(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %snick <new-nick>\n",
				     command_prefix);
		return;
	}

	send_enaml_buildf(ci->server->sock, "change-nick", "nick:%s", text);
}

/* user_whom
 * Handle the /whom command
 */
static void
user_whom(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (strlen(text)) {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s verbose grep:%s", "user", text);
	} else {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s verbose", "user");
	}
}

/* user_who
 * Handle the /who command
 */
static void
user_who(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (strlen(text)) {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s grep:%s", "user", text);
	} else {
		send_enaml_buildf(ci->server->sock, "list-clients",
				  "type:%s", "user");
	}
}

/* user_whois
 * Handle the /whois command
 */
static void
user_whois(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %swhois <nick>\n",
				     command_prefix);
		return;
	}

	send_enaml_buildf(ci->server->sock, "whois", "nick:%s type:%s",
			  text, "user");
}

/* user_seen
 * Handle the /seen command
 */
static void
user_seen(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %sseen <nick>\n",
				     command_prefix);
		return;
	}

	send_enaml_buildf(ci->server->sock, "whois", "id:%s type:%s",
			  text, "user");
}

/* user_channel
 * Handle the /channel command
 */
static void
user_channel(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %schannel <nick>\n",
				     command_prefix);
		return;
	}

	send_enaml_buildf(ci->server->sock, "whois", "id:%s type:%s",
			  text, "channel");
}

/* user_say
 * Handle the /say command
 */
static void
user_say(ChatInterface *ci, const gchar *text)
{
	gchar *text2;

	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %ssay <text>\n",
				     command_prefix);
		return;
	}

	if (ci->server->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	text2 = utf8_encode(text);
	send_enaml_buildf(ci->server->sock, "say", "text:%s", text2);
	g_free(text2);
}

/* user_me
 * Handle the /me command
 */
static void
user_me(ChatInterface *ci, const gchar *text)
{
	gchar *text2;
	
	g_assert(ci);

	/* If the Recipient is a USER, /me actually does a /act */
	if (ci->recip_type == CHAT_RECIPTYPE_USER) {
		gchar *msg;

		msg = g_strdup_printf("%s %s", ci->recip, text);
		user_act(ci, msg);
		g_free(msg);

		return;
	}

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %sme <text>\n",
				     command_prefix);
		return;
	}

	if (ci->server->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	text2 = utf8_encode(text);
	send_enaml_buildf(ci->server->sock, "say", "act:%s", text2);
	g_free(text2);
}

/* user_msg
 * Handle the /msg command
 */
static void
user_msg(ChatInterface *ci, const gchar *text)
{
	gchar **arg;
	gchar  *text2;

	g_assert(ci);

	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** Usage: %smsg <nick> <text>\n",
				     command_prefix);
		return;
	}

	if (ci->server->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	text2 = utf8_encode(arg[1]);
	send_enaml_buildf(ci->server->sock, "message", "nick:%s text:%s",
			  arg[0], text2);
	g_free(text2);

	g_strfreev(arg);
}

/* user_act
 * Handle the /act command
 */
static void
user_act(ChatInterface *ci, const gchar *text)
{
	gchar **arg;
	gchar  *text2;

	g_assert(ci);

	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0]) || !arg[1] || !strlen(arg[1])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** Usage: %sact <nick> <text>\n",
				     command_prefix);
		return;
	}

	if (ci->server->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	text2 = utf8_encode(arg[1]);
	send_enaml_buildf(ci->server->sock, "message", "nick:%s act:%s",
			  arg[0], text2);
	g_free(text2);

	g_strfreev(arg);
}

/* user_wall
 * Handle the /wall command
 */
static void
user_wall(ChatInterface *ci, const gchar *text)
{
	gchar *text2;

	g_assert(ci);

	if (!strlen(text)) {
		chat_interface_sayto(ci, "*** Usage: %swall <text>\n",
				     command_prefix);
		return;
	}

	if (ci->server->status == SAY2_STATUS_AWAY) {
		send_enaml_buildf(ci->server->sock, "status", "mode:%s text:%s",
				  "active", "");
	}

	text2 = utf8_encode(text);
	send_enaml_buildf(ci->server->sock, "wall", "text:%s", text2);
	g_free(text2);
}

/* user_history
 * Handle the /history command
 */
static void
user_history(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);

	if (strlen(text)) {
		send_enaml_buildf(ci->server->sock, "history", "max:%s", text);
	} else {
		send_enaml_buildf(ci->server->sock, "history", "");
	}
}

/* user_boot
 * Handle the /boot command
 */
static void
user_boot(ChatInterface *ci, const gchar *text)
{
	gchar **arg;

	g_assert(ci);

	arg = g_strsplit(text, " ", 1);
	if (!arg[0] || !strlen(arg[0])) {
		g_strfreev(arg);
		chat_interface_sayto(ci, "*** Usage: %sboot <nick> [reason]\n",
				     command_prefix);
		return;
	}

	if (arg[1] && strlen(arg[1])) {
		gchar *text2;

		text2 = utf8_encode(arg[1]);
		send_enaml_buildf(ci->server->sock, "boot-user",
				  "nick:%s text:%s", arg[0], text2);
		g_free(text2);
	} else {
		send_enaml_buildf(ci->server->sock, "boot-user", "nick:%s",
				  arg[0]);
	}
}

/* user_clear
 * Handle the /clear command
 */
static void
user_clear(ChatInterface *ci, const gchar *text)
{
	g_assert(ci);
	gtk_editable_delete_text(GTK_EDITABLE(ci->output_box), 0, -1);
}


/* chat_interface_entry_activated
 * Called when they type text into the entry box and press enter
 */
static void
chat_interface_entry_activated(GtkWidget *widget, ChatInterface *ci)
{
	static void *dispatch_table = NULL;
	static void *uc_dispatch_table = NULL;
	const gchar *text;

	g_assert(widget);
	g_assert(ci);

	/* Oh god, I'm turning into poskanser */
	if (!dispatch_table)
		dispatch_table = dispatch_create(command_table);
	if (!uc_dispatch_table)
		uc_dispatch_table = dispatch_create(uc_command_table);

	/* Get the text they typed into the widget */
	text = gtk_entry_get_text(GTK_ENTRY(widget));

	/* Mediumly kludgy, but it'll work */
	if (!strncmp(text, command_prefix, strlen(command_prefix))) {
		/* Command */
		void        *table;
		gchar       *cmd;
		size_t       cmdlen;

		/* Copy the command out */
		text += strlen(command_prefix);
		cmdlen = strcspn(text, " ");
		cmd = g_strndup(text, cmdlen);

		/* Find the start of the text */
		text += cmdlen;
		if (*text)
			text++;
		
		/* Choose a dispatch table */
		table = (ci->server && ci->server->sock ? dispatch_table
			 :  uc_dispatch_table);

		/* Try to dispatch it */
		if (!dispatch_go(table, cmd, ci, (void *)text, NULL)) {
			chat_interface_say(ci->server, "*** Unknown "
					               "command: %s\n", cmd);
		}

	} else if (ci->recip_type == CHAT_RECIPTYPE_CHAN) {
		/* Say */
		if (ci->server && ci->server->sock) {
			user_say(ci, text);
		} else {
			chat_interface_say(ci->server, "*** Not connected\n");
		}

	} else if (ci->recip_type == CHAT_RECIPTYPE_USER) {
		/* Msg */
		if (ci->server && ci->server->sock) {
			gchar *msg;

			msg = g_strdup_printf("%s %s", ci->recip, text);
			user_msg(ci, msg);
			g_free(msg);
		} else {
			chat_interface_say(ci->server, "*** Not connected\n");
		}
	}

	/* Clear the widget */
	gtk_entry_set_text(GTK_ENTRY(widget), "");
}

/* chat_interface_saytov
 * Output formatted information to a specific chat interface
 */
void
chat_interface_saytov(ChatInterface *ci, const gchar *msg, va_list args)
{
	gchar *text;

	g_assert(ci);

	text = g_strdup_vprintf(msg, args);

	gtk_text_insert(GTK_TEXT(ci->output_box), NULL, NULL, NULL,
			text, strlen(text));

	g_free(text);
}

/* chat_interface_sayv
 * Output formatted information to the right chat interface(s)
 */
void
chat_interface_sayv(Say2Server *server, const gchar *msg, va_list args)
{
	GSList *list;

	for (list = open_interfaces; list; list = g_slist_next(list)) {
		ChatInterface *ci;

		ci = list->data;
		g_assert(ci);

		if (ci->server != server)
			continue;

		chat_interface_saytov(ci, msg, args);
	}
}

/* chat_interface_sayto
 * Output formatted information to a specific chat interface
 */
void
chat_interface_sayto(ChatInterface *ci, const gchar *msg, ...)
{
	va_list args;

	va_start(args, msg);
	chat_interface_saytov(ci, msg, args);
	va_end(args);
}

/* chat_interface_say
 * Output formatted information to the right chat interface(s)
 */
void
chat_interface_say(Say2Server *server, const gchar *msg, ...)
{
	va_list args;

	va_start(args, msg);
	chat_interface_sayv(server, msg, args);
	va_end(args);
}

/* chat_interface_update
 * Update the labels and text that contain things like the nickname
 */
void
chat_interface_update(ChatInterface *ci)
{
	g_assert(ci);
	g_assert(ci->server);

	g_return_if_fail(ci->server->nickname);
	g_return_if_fail(ci->server->server_name);

	gtk_label_set_text(GTK_LABEL(ci->nick_label),
			   ci->server->nickname);

	if (ci->title_win) {
		GString *title;

		title = g_string_new(ci->title_prefix);
		g_string_append(title, ci->server->nickname);
		g_string_append(title, " @ ");
		g_string_append(title, ci->server->server_name);
		if (ci->server->server_port != 3797)
			g_string_sprintfa(title, ":%d",
					  ci->server->server_port);

		gtk_window_set_title(GTK_WINDOW(ci->title_win), title->str);

		g_string_free(title, TRUE);
	}

	gtk_widget_grab_focus(ci->entry_box);
}

/* chat_interface_update_by_server
 * For all chat interfaces with the server given, update labels and text
 */
void
chat_interface_update_by_server(Say2Server *server)
{
	GSList *list;

	g_assert(server);

	for (list = open_interfaces; list; list = g_slist_next(list)) {
		ChatInterface *ci;

		ci = list->data;
		g_assert(ci);

		if (ci->server != server)
			continue;

		chat_interface_update(ci);
	}
}

/* chat_interface_set_destination
 * Set the destination for non-prefixed text typed into the entry box
 */
void
chat_interface_set_destination(ChatInterface *ci, ChatRecipType recip_type,
		               const gchar *recip)
{
	g_assert(ci);

	ci->recip_type = recip_type;
	if (ci->recip_type == CHAT_RECIPTYPE_ANY)
		ci->recip_type = CHAT_RECIPTYPE_NONE;

	g_free(ci->recip);
	ci->recip = g_strdup(recip);
}

/* chat_interface_set_title_win
 * Set the window to change the title of
 */
void
chat_interface_set_title_win(ChatInterface *ci, GtkWidget *title_win,
		             const gchar *title_prefix)
{
	g_assert(ci);

	ci->title_win = title_win;

	g_free(ci->title_prefix);
	ci->title_prefix = g_strdup(title_prefix);
}

/* chat_interface_set_server
 * Set the Say2Server we use to communicate on
 */
void
chat_interface_set_server(ChatInterface *ci, Say2Server *server)
{
	/* Decrease the ref count of the current server and free if need to */
	if (ci->server) {
		ci->server->ref--;
		if (!ci->server->ref)
			say2_server_free(ci->server);
	}

	/* Server == 1 means create one and ask for connection details */
	if (server == (Say2Server *)1) {
		ci->server = say2_server_new();
		ci->server->ref++;
		connect_dialog_new(NULL, (ConnectDialogFn)&say2_server_connect,
				   ci->server);
	} else if (server) {
		/* Increase the ref count */
		ci->server = server;
		if (ci->server)
			ci->server->ref++;
	} else {
		/* Assigning NULL is acceptable */
		ci->server = NULL;
	}
}

/* chat_interface_free
 * Free up the data used by a chat interface
 */
void
chat_interface_free(ChatInterface *ci)
{
	GSList *list;

	if (!ci)
		return;

	/* Remove the server from it, and remove it from the list */
	chat_interface_set_server(ci, NULL);
	open_interfaces = g_slist_remove(open_interfaces, ci);

	/* Free all the widgets */
	gtk_widget_destroy(ci->top_vbox);

	/* Free the list of things we accept */
	for (list = ci->accept_list; list; list = g_slist_next(list)) {
		ChatAccept *ca;

		ca = list->data;
		g_assert(ca);

		g_free(ca->recip);
		g_free(ca);

		list->data = NULL;
	}
	g_slist_free(ci->accept_list);

	g_free(ci->recip);
	g_free(ci);
}

/* chat_interface_new
 * Create a new chat interface, returns the structure of information for
 * storage or futher manipulation
 */
ChatInterface *
chat_interface_new(void)
{
	ChatInterface *ci;
	GtkWidget     *hbox, *scroll;
	GdkFont       *chat_font;
	GtkStyle      *style;

	/* Load the font */
	chat_font = gdk_font_load(CHAT_FONT);

	/* Create the interface and add it to the list */
	ci = g_new0(ChatInterface, 1);
	open_interfaces = g_slist_append(open_interfaces, ci);

	/* Create a box for the elements */
	ci->top_vbox = gtk_vbox_new(FALSE, 4);
	gtk_widget_show(ci->top_vbox);

	/* Create a box for the output widget and its scrollbar */
	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(ci->top_vbox), hbox, TRUE, TRUE, 0);
	gtk_widget_show(hbox);

	/* Create a text box for the output */
	ci->output_box = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(ci->output_box), FALSE);
	gtk_box_pack_start(GTK_BOX(hbox), ci->output_box, TRUE, TRUE, 0);
	gtk_widget_show(ci->output_box);

	/* Restyle the text box */
	style = gtk_style_copy(gtk_widget_get_style(ci->output_box));
	style->font = chat_font;
	gtk_widget_set_style(ci->output_box, style);

	/* Create scroll bar for the text box */
	scroll = gtk_vscrollbar_new(GTK_TEXT(ci->output_box)->vadj);
	gtk_box_pack_end(GTK_BOX(hbox), scroll, FALSE, FALSE, 0);
	gtk_widget_show(scroll);

	/* Create a box for the nick label and the entry box */
	hbox = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_end(GTK_BOX(ci->top_vbox), hbox, FALSE, FALSE, 0);
	gtk_widget_show(hbox);

	/* Create a label for the nickname */
	ci->nick_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(hbox), ci->nick_label, FALSE, FALSE, 0);
	gtk_widget_show(ci->nick_label);

	/* Create an entry box */
	ci->entry_box = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(ci->entry_box), "activate",
			   GTK_SIGNAL_FUNC(chat_interface_entry_activated),
			   (gpointer)ci);
	gtk_box_pack_end(GTK_BOX(hbox), ci->entry_box, TRUE, TRUE, 0);
	gtk_widget_show(ci->entry_box);

	/* Restyle the entry box */
	style = gtk_style_copy(gtk_widget_get_style(ci->entry_box));
	style->font = chat_font;
	gtk_widget_set_style(ci->entry_box, style);

	return ci;
}
