/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * common_dialog.c
 *  - The bit of code that is common to creation of all dialogs.
 * -- 
 * $Id: common_dialog.c,v 1.1 2001/06/04 11:55:42 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <stdarg.h>

#include <gtk/gtk.h>

#include <meko.h>

#include "common_dialog.h"

/* common_dialog_new
 * Wrapper around common_dialog_newv
 */
GtkWidget *
common_dialog_new(const gchar *title, const gchar *wmclass_name, ...)
{
	GtkWidget *dialog;
	va_list    args;

	va_start(args, wmclass_name);
	dialog = common_dialog_newv(title, wmclass_name, args);
	va_end(args);

	return dialog;
}

/* common_dialog_newv
 * Create a common dialog template.  The extra arguments are a list of buttons
 * for the dialog, as follows:
 *     const gchar    *label
 *     GtkSignalFunc   callback
 *     gpointer        data
 *     GtkObject      *slot_object
 *     GtkWidget     **widget_ptr
 *     gboolean        default_action
 */
GtkWidget *
common_dialog_newv(const gchar *title, const gchar *wmclass_name, va_list args)
{
	GtkWidget *dialog, *action_area;

	/* Create the dialog */
	dialog = gtk_dialog_new();

	/* Set up the dialog basics */
	if (wmclass_name) {
		gtk_window_set_wmclass(GTK_WINDOW(dialog),
				       wmclass_name, "Meko");
	} else {
		gtk_window_set_wmclass(GTK_WINDOW(dialog), "meko", "Meko");
	}
	if (title) {
		gtk_window_set_title(GTK_WINDOW(dialog), title);
	} else {
		gtk_window_set_title(GTK_WINDOW(dialog), "Meko");
	}
	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, TRUE);

	/* Prepare the action area */
	action_area = GTK_DIALOG(dialog)->action_area;
	gtk_container_set_border_width(GTK_CONTAINER(action_area), 2);
	gtk_box_set_homogeneous(GTK_BOX(action_area), FALSE);

	/* Add action area buttons */
	common_dialog_buttonsv(dialog, action_area, args);

	return dialog;
}

/* common_dialog_buttons
 * Wrapper around common_dialog_buttonsv
 */
GtkWidget *
common_dialog_buttons(GtkWidget *dialog, GtkWidget *action_area, ...)
{
	GtkWidget *bbox;
	va_list    args;

	va_start(args, action_area);
	bbox = common_dialog_buttonsv(dialog, action_area, args);
	va_end(args);

	return bbox;
}

/* common_dialog_buttonsv
 * Create a button box with lots of buttons in it for a dialog.
 *     const gchar    *label
 *     GtkSignalFunc   callback
 *     gpointer        data
 *     GtkObject      *slot_object
 *     GtkWidget     **widget_ptr
 *     gboolean        default_action
 */
GtkWidget *
common_dialog_buttonsv(GtkWidget *dialog, GtkWidget *action_area, va_list args)
{
	const gchar *label;
	GtkWidget   *bbox;

	/* Create a button box for the buttons */
	bbox = gtk_hbutton_box_new();
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 4);
	gtk_widget_show(bbox);

	/* Pack it into the end of an action_area if one is given */
	if (action_area)
		gtk_box_pack_end(GTK_BOX(action_area), bbox, FALSE, FALSE, 0);

	/* Process the buttons */
	label = va_arg(args, const gchar *);
	while (label) {
		GtkWidget      *button;
		GtkObject      *slot_object;
		GtkWidget     **widget_ptr;
		GtkSignalFunc   callback;
		gboolean        default_action;
		gpointer        data;

		/* Get the rest of the arguments */
		callback = va_arg(args, GtkSignalFunc);
		data = va_arg(args, gpointer);
		slot_object = va_arg(args, GtkObject *);
		widget_ptr = va_arg(args, GtkWidget **);
		default_action = va_arg(args, gboolean);

		/* Create the button */
		button = gtk_button_new_with_label(label);
		GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(bbox), button, FALSE, FALSE, 0);

		/* A slot_object of 1 means the dialog itself */
		if (slot_object == (GtkObject *)1) {
			if (dialog)
				slot_object = GTK_OBJECT(dialog);
			else
				slot_object = NULL;
		}

		/* Connect the callback */
		if (callback) {
			if (slot_object) {
				gtk_signal_connect_object(
					GTK_OBJECT(button), "clicked",
					GTK_SIGNAL_FUNC(callback), slot_object);
			} else {
				gtk_signal_connect(
					GTK_OBJECT(button), "clicked",
					GTK_SIGNAL_FUNC(callback), data);
			}
		}

		/* Give them the widget back if they want it */
		if (widget_ptr)
			*widget_ptr = button;

		/* Do default and show button */
		if (default_action)
			gtk_widget_grab_default(button);
		gtk_widget_show(button);

		/* Get the next label if there is one */
		label = va_arg(args, const gchar *);
	}

	return bbox;
}
