/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * connect_dialog.c
 *  - Dialog to ask user for connection information
 * -- 
 * $Id: connect_dialog.c,v 1.4 2001/06/05 11:22:10 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <gtk/gtk.h>

#include <meko.h>

#include "connect_dialog.h"
#include "connect_values.h"
#include "common_dialog.h"

/* structure to hold pointers to the various fields of a connection dialog
 * that we need to carry about
 */
typedef struct _connect_dialog ConnectDialog;
struct _connect_dialog {
	GtkWidget       *dialog;
	ConnectValues   *values;
	ConnectDialogFn  callback;
	void            *callback_ptr;

	/* Fields */
	GtkWidget *server_entry;
	GtkWidget *port_entry;
	GtkWidget *name_entry;
	GtkWidget *user_radio;
	GtkWidget *guest_radio;
	GtkWidget *pass_entry;
	GtkWidget *mail_entry;
	GtkWidget *tag_entry;

	/* Need to keep these so we can enable/disable them */
	GtkWidget *pass_label;
	GtkWidget *mail_label;
	GtkWidget *tag_label;
};

/* forward prototypes */
static void connect_dialog_required_destroy(GtkWidget *, GtkWidget *);
static void connect_dialog_required(ConnectDialog *,
		                    const gchar *, GtkWidget *);
static void connect_dialog_ok(GtkWidget *, ConnectDialog *);
static void connect_dialog_login_toggled(GtkWidget *, ConnectDialog *);
static void connect_dialog_reset(GtkWidget *, ConnectDialog *);
static void connect_dialog_destroy(GtkWidget *, ConnectDialog *);

/* Currently open connection dialog.  I see no reason to allow more than one
 * of these at a time, even if they're for different windows/tabs - it's only
 * confusing
 */
static GtkWidget *current_dialog = NULL;


/* connect_dialog_ok
 * Called when the user OKs the dialog
 */
static void
connect_dialog_ok(GtkWidget *widget, ConnectDialog *d)
{
	const gchar *server_name, *server_port;
	const gchar *nickname, *password, *email, *tagline;
	gboolean     as_guest;

	g_assert(d);

	server_name = gtk_entry_get_text(GTK_ENTRY(d->server_entry));
	server_port = gtk_entry_get_text(GTK_ENTRY(d->port_entry));
	nickname = gtk_entry_get_text(GTK_ENTRY(d->name_entry));
	password = gtk_entry_get_text(GTK_ENTRY(d->pass_entry));
	email = gtk_entry_get_text(GTK_ENTRY(d->mail_entry));
	tagline = gtk_entry_get_text(GTK_ENTRY(d->tag_entry));
	
	as_guest = gtk_toggle_button_get_active(
			GTK_TOGGLE_BUTTON(d->guest_radio));

	/* If any required field is empty, let them know */
	if (!strlen(server_name)) {
		connect_dialog_required(d, "Server", d->server_entry);
	} else if (!strlen(server_port)) {
		connect_dialog_required(d, "Port", d->port_entry);
	} else if (!strlen(nickname)) {
		connect_dialog_required(d, "Nickname", d->name_entry);
	} else if (!strlen(password) && !as_guest) {
		connect_dialog_required(d, "Password", d->pass_entry);
	} else if (!strlen(email) && as_guest) {
		connect_dialog_required(d, "E-Mail", d->mail_entry);
	} else {
		/* Save the values.  Can't just point at the text boxes
		 * because otherwise we'll free them twice (boom!)
		 */
		connect_values_free(d->values);
		d->values = connect_values_new();
		d->values->server_name = g_strdup(server_name);
		d->values->server_port = g_strdup(server_port);
		d->values->nickname = g_strdup(nickname);
		d->values->password = g_strdup(password);
		d->values->email = g_strdup(email);
		d->values->tagline = g_strdup(tagline);
		d->values->as_guest = as_guest;
	
		d->callback(d->values, d->callback_ptr);
		gtk_widget_destroy(d->dialog);
	}
}

/* connect_dialog_login_toggled
 * Called whenever the user toggles either of the "Login as" radio buttons.
 * GTK+ will actually call this once for each member of the group, so we can
 * neaten the code by hooking it the right way.
 */
static void
connect_dialog_login_toggled(GtkWidget *widget, ConnectDialog *d)
{
	gboolean active;

	g_assert(d);

	active = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));

	if (widget == d->user_radio) {
		gtk_widget_set_sensitive(d->pass_label, active);
		gtk_widget_set_sensitive(d->pass_entry, active);
	} else if (widget == d->guest_radio) {
		gtk_widget_set_sensitive(d->mail_label, active);
		gtk_widget_set_sensitive(d->mail_entry, active);
		gtk_widget_set_sensitive(d->tag_label, active);
		gtk_widget_set_sensitive(d->tag_entry, active);
	}
}

/* connect_dialog_reset
 * Reset a connection dialog back to defaults.
 */
static void
connect_dialog_reset(GtkWidget *widget, ConnectDialog *d)
{
	g_assert(d);

	gtk_entry_set_text(GTK_ENTRY(d->server_entry), d->values->server_name);
	gtk_entry_set_text(GTK_ENTRY(d->port_entry), d->values->server_port);
	gtk_entry_set_text(GTK_ENTRY(d->name_entry), d->values->nickname);
	gtk_entry_set_text(GTK_ENTRY(d->pass_entry), d->values->password);
	gtk_entry_set_text(GTK_ENTRY(d->mail_entry), d->values->email);
	gtk_entry_set_text(GTK_ENTRY(d->tag_entry), d->values->tagline);

	/* Will inherently turn off the other one */
	if (d->values->as_guest) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->guest_radio),
				             TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(d->user_radio),
				             TRUE);
	}
}

/* connect_dialog_destroy
 * Make sure that the structure containing all the pointers is destroyed
 * when the window is
 */
static void
connect_dialog_destroy(GtkWidget *widget, ConnectDialog *d)
{
	g_assert(d);

	/* Free up the connect_dialog structure that goes with this one */
	connect_values_free(d->values);
	g_free(d);

	/* Let them open another one */
	current_dialog = NULL;
}

/* connect_dialog_required_destroy
 * Called when a field required error box is closed.  Reactivates the parent
 * dialog and gives focus to the right widget
 */
static void
connect_dialog_required_destroy(GtkWidget *dialog, GtkWidget *field_widget)
{
	ConnectDialog *d;

	g_assert(field_widget);

	/* Get the parent dialog */
	d = (ConnectDialog *)gtk_object_get_user_data(GTK_OBJECT(dialog));
	g_assert(d);

	/* Make the original dialog enabled again */
	gtk_widget_set_sensitive(d->dialog, TRUE);

	/* Focus the widget that caused the error */
	gtk_widget_grab_focus(field_widget);
}

/* connect_dialog_required
 * Display a simple error box indicating a field was required
 */
static void
connect_dialog_required(ConnectDialog *d, const gchar *field_name,
		        GtkWidget *field_widget)
{
	GtkWidget *dialog, *vbox, *label;
	gchar     *message;

	g_assert(d);
	g_assert(field_name);
	g_assert(field_widget);
	
	/* Make the original dialog disabled */
	gtk_widget_set_sensitive(d->dialog, FALSE);

	/* Create ourselves a nice dialog */
	dialog = common_dialog_new("Error", "error",
			           "OK",
				   GTK_SIGNAL_FUNC(gtk_widget_destroy),
				   NULL, (GtkObject *)1, NULL, TRUE,
				   NULL);
	gtk_object_set_user_data(GTK_OBJECT(dialog), (gpointer)d);
	gtk_signal_connect(GTK_OBJECT(dialog), "destroy",
			   GTK_SIGNAL_FUNC(connect_dialog_required_destroy),
			   (gpointer)field_widget);

	/* Create a box inside the dialog */
	vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vbox,
			   TRUE, TRUE, 0);
	gtk_widget_show(vbox);

	/* Create a label */
	message = g_strdup_printf("The '%s' field is required.", field_name);
	label = gtk_label_new(message);
	gtk_misc_set_alignment(GTK_MISC(label), 0.5, 0.5);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	/* Show the dialog */
	gtk_widget_show(dialog);
}

/* connect_dialog_new
 * Create and display a New Connection dialog
 */
void
connect_dialog_new(ConnectValues *orig_values, ConnectDialogFn callback,
		   void *callback_ptr)
{
	GtkWidget     *top_vbox, *frame, *box, *table, *w;
	GSList        *group;
	ConnectDialog *d;

	g_assert(callback);

	/* Focus a current dialog, don't open a new one */
	if (current_dialog) {
		gtk_widget_grab_focus(current_dialog);
		return;
	}

	/* Create the dialog */
	d = g_new0(ConnectDialog, 1);
	d->callback = callback;
	d->callback_ptr = callback_ptr;
	d->values = connect_values_new_from(orig_values);
	d->dialog = common_dialog_new("New Connection", "new_connection",
			              "OK",
				      GTK_SIGNAL_FUNC(connect_dialog_ok),
				      (gpointer)d, NULL, NULL, TRUE,
			              "Reset",
				      GTK_SIGNAL_FUNC(connect_dialog_reset),
				      (gpointer)d, NULL, NULL, FALSE,
			              "Cancel",
				      GTK_SIGNAL_FUNC(gtk_widget_destroy),
				      NULL, (GtkObject *)1, NULL, FALSE,
				      NULL);
	gtk_signal_connect(GTK_OBJECT(d->dialog), "destroy",
			   GTK_SIGNAL_FUNC(connect_dialog_destroy),
			   (gpointer)d);
	current_dialog = d->dialog;

	/* Box for everything else */
	top_vbox = gtk_vbox_new(FALSE, 4);
	gtk_container_set_border_width(GTK_CONTAINER(top_vbox), 4);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(d->dialog)->vbox), top_vbox,
			   TRUE, TRUE, 0);
	gtk_widget_show(top_vbox);

	/* Frame for server and port */
	frame = gtk_frame_new("Connect To:");
	gtk_box_pack_start(GTK_BOX(top_vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	/* Packing box for Server and Port entry boxes */
	box = gtk_hbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(box), 2);
	gtk_container_add(GTK_CONTAINER(frame), box);
	gtk_widget_show(box);

	/* Server label */
	w = gtk_label_new("Server:");
	gtk_box_pack_start(GTK_BOX(box), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Server entry box */
	d->server_entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(box), d->server_entry, FALSE, FALSE, 0);
	gtk_widget_show(d->server_entry);

	/* Port label (extra space to seperate it a bit from the server) */
	w = gtk_label_new(" Port:");
	gtk_box_pack_start(GTK_BOX(box), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Port entry box */
	d->port_entry = gtk_entry_new();
	gtk_widget_set_usize(d->port_entry, 40, -1);
	gtk_box_pack_start(GTK_BOX(box), d->port_entry, FALSE, FALSE, 0);
	gtk_widget_show(d->port_entry);
	
	/* Frame for login as */
	frame = gtk_frame_new("Connect As:");
	gtk_box_pack_start(GTK_BOX(top_vbox), frame, FALSE, FALSE, 0);
	gtk_widget_show(frame);

	/* Table for everything in the frame */
	table = gtk_table_new(6, 3, FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(table), 2);
	gtk_container_add(GTK_CONTAINER(frame), table);
	gtk_widget_show(table);

	/* Packing box for nickname */
	box = gtk_hbox_new(FALSE, 0);
	gtk_table_attach(GTK_TABLE(table), box, 0, 3, 0, 1,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(box);

	/* Nickname label */
	w = gtk_label_new("Nickname:");
	gtk_box_pack_start(GTK_BOX(box), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Nickname entry box */
	d->name_entry = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(box), d->name_entry, FALSE, FALSE, 0);
	gtk_widget_show(d->name_entry);

	/* Separator */
	w = gtk_hseparator_new();
	gtk_table_attach(GTK_TABLE(table), w, 0, 3, 1, 2,
			 GTK_EXPAND | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 2);
	gtk_widget_show(w);

	/* Login as registered user button */
	d->user_radio = gtk_radio_button_new_with_label(NULL, "Login as "
			                                "Registered User");
	group = gtk_radio_button_group(GTK_RADIO_BUTTON(d->user_radio));
	gtk_signal_connect(GTK_OBJECT(d->user_radio), "toggled",
			   GTK_SIGNAL_FUNC(connect_dialog_login_toggled),
			   (gpointer)d);
	gtk_table_attach(GTK_TABLE(table), d->user_radio, 0, 3, 2, 3,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->user_radio);

	/* Login as guest button */
	d->guest_radio = gtk_radio_button_new_with_label(group,
			                                 "Login as Guest");
	gtk_signal_connect(GTK_OBJECT(d->guest_radio), "toggled",
			   GTK_SIGNAL_FUNC(connect_dialog_login_toggled),
			   (gpointer)d);
	gtk_table_attach(GTK_TABLE(table), d->guest_radio, 0, 3, 4, 5,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->guest_radio);

	/* Password label */
	d->pass_label = gtk_label_new("Password:");
	gtk_misc_set_alignment(GTK_MISC(d->pass_label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), d->pass_label, 1, 2, 3, 4,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->pass_label);

	/* Password entry box */
	d->pass_entry = gtk_entry_new();
	gtk_entry_set_visibility(GTK_ENTRY(d->pass_entry), FALSE);
	gtk_table_attach(GTK_TABLE(table), d->pass_entry, 2, 3, 3, 4,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->pass_entry);

	/* E-Mail Address label */
	d->mail_label = gtk_label_new("E-Mail:");
	gtk_misc_set_alignment(GTK_MISC(d->mail_label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), d->mail_label, 1, 2, 5, 6,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->mail_label);

	/* E-Mail Address entry box */
	d->mail_entry = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), d->mail_entry, 2, 3, 5, 6,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->mail_entry);

	/* Tag Line label */
	d->tag_label = gtk_label_new("Tag Line:");
	gtk_misc_set_alignment(GTK_MISC(d->tag_label), 1.0, 0.5);
	gtk_table_attach(GTK_TABLE(table), d->tag_label, 1, 2, 6, 7,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->tag_label);

	/* Tag Line entry box */
	d->tag_entry = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), d->tag_entry, 2, 3, 6, 7,
			 GTK_SHRINK | GTK_FILL, GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(d->tag_entry);

	/* Click the reset button */
	connect_dialog_reset(d->dialog, d);
	gtk_widget_set_sensitive(d->pass_label, !d->values->as_guest);
	gtk_widget_set_sensitive(d->pass_entry, !d->values->as_guest);
	gtk_widget_set_sensitive(d->mail_label, d->values->as_guest);
	gtk_widget_set_sensitive(d->mail_entry, d->values->as_guest);
	gtk_widget_set_sensitive(d->tag_label, d->values->as_guest);
	gtk_widget_set_sensitive(d->tag_entry, d->values->as_guest);

	/* Show the dialog */
	gtk_widget_show(d->dialog);
	gtk_widget_grab_focus(d->server_entry);
}
