/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * say2_server.c
 *  - Communication with a say2 server
 * -- 
 * $Id: say2_server.c,v 1.7 2001/06/07 12:35:38 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <glib.h>
#include <say2/say2.h>

#include <meko.h>

#include "say2_server.h"
#include "connect_values.h"
#include "chat_interface.h"
#include "whois_info.h"
#include "unicode.h"

/* Forward prototypes for command functions*/
static void do_error(Say2Server *, const EnamlBlock *);
static void do_hello(Say2Server *, const EnamlBlock *);
static void do_server_info(Say2Server *, const EnamlBlock *);
static void do_motd(Say2Server *, const EnamlBlock *);
static void do_welcome(Say2Server *, const EnamlBlock *);
static void do_bad_nick(Say2Server *, const EnamlBlock *);
static void do_ping(Say2Server *, const EnamlBlock *);
static void do_signon(Say2Server *, const EnamlBlock *);
static void do_signoff(Say2Server *, const EnamlBlock *);
static void do_prefs(Say2Server *, const EnamlBlock *);
static void do_status(Say2Server *, const EnamlBlock *);
static void do_change_nick(Say2Server *, const EnamlBlock *);
static void do_members(Say2Server *, const EnamlBlock *);
static void do_begin_response(Say2Server *, const EnamlBlock *);
static void do_end_response(Say2Server *, const EnamlBlock *);
static void do_client_info(Say2Server *, const EnamlBlock *);
static void do_not_present(Say2Server *, const EnamlBlock *);
static void do_no_such_client(Say2Server *, const EnamlBlock *);
#if 0
static void do_client_attributes(Say2Server *, const EnamlBlock *);
static void do_set_attributes(Say2Server *, const EnamlBlock *);
#endif
static void do_say(Say2Server *, const EnamlBlock *);
static void do_message(Say2Server *, const EnamlBlock *);
static void do_wall(Say2Server *, const EnamlBlock *);

/* Forward prototypes for other function */
static void say2_server_read(Socket *);
static void say2_server_ready(Socket *, gint);
static void say2_server_eof(Socket *);
static void say2_server_free_values(Say2Server *);

/* Command dispatch table */
static DispatchItem command_table[] = {
	{ "error", &do_error },
	{ "hello", &do_hello },
	{ "server-info", &do_server_info },
	{ "motd", &do_motd },
	{ "welcome", &do_welcome },
	{ "bad-nick", &do_bad_nick },
	{ "ping", &do_ping },
	{ "signon", &do_signon },
	{ "signoff", &do_signoff },
	{ "prefs", &do_prefs },
	{ "status", &do_status },
	{ "change-nick", &do_change_nick },
	{ "members", &do_members },
	{ "begin-response", &do_begin_response },
	{ "end-response", &do_end_response },
	{ "client-info", &do_client_info },
	{ "not-present", &do_not_present },
	{ "no-such-client", &do_no_such_client },
#if 0
	{ "client-attributes", &do_client_attributes },
	{ "set-attributes", &do_set_attributes },
#endif
	{ "say", &do_say },
	{ "message", &do_message },
	{ "wall", &do_wall },
	{ NULL, NULL }
};


/* do_error
 * Handle error command
 */
static void
do_error(Say2Server *server, const EnamlBlock *error)
{
	const gchar *command;
	gchar       *text;

	g_assert(server);

	if (!error) {
		chat_interface_say(server, "*** Error\n");
		return;
	}

	command = enaml_get_string(error, "command", NULL);
	text = utf8_decode(enaml_get_string(error, "text", ""));

	if (command) {
		chat_interface_say(server, "*** %s Error: %s\n",
				   command, text);
	} else {
		chat_interface_say(server, "*** Error: %s\n", text);
	}
	
	g_free(text);
}

/* do_hello
 * Handle hello command
 */
static void
do_hello(Say2Server *server, const EnamlBlock *hello)
{
	EnamlBlock *login, *block;

	g_assert(server);
	
	/* Construct login block */
	login = enaml_new();
	enaml_add_string(login, "client", "Meko " FULL_VERSION, ENAML_COPY);
	enaml_add_string(login, "nick", server->nickname, ENAML_COPY);
	if (server->password) {
		enaml_add_string(login, "password", server->password,
				 ENAML_COPY);
	} else {
		enaml_add_string(login, "email", server->email, ENAML_COPY);
		if (server->tagline) {
			gchar *tagline;

			tagline = utf8_encode(server->tagline);
			enaml_add_string(login, "tagline", tagline, ENAML_COPY);
			g_free(tagline);
		}
	}

	/* Construct block to send */
	block = enaml_new();
	enaml_add_block(block, "login", login, ENAML_FREE);
	send_enaml_block(server->sock, block);
	enaml_free(block);
}

/* do_server_info
 * Handle server-info command
 */
static void
do_server_info(Say2Server *server, const EnamlBlock *server_info)
{
	const gchar *server_prog, *version;
	gchar       *text, *uptime;

	g_assert(server);

	if (!server_info) {
		chat_interface_say(server, "*** Bogus server-info received\n");
		return;
	}

	server_prog = enaml_get_string(server_info, "server", "?");
	version = enaml_get_string(server_info, "version", "?");
	text = utf8_decode(enaml_get_string(server_info, "text", NULL));
	uptime = whois_timespan_str(enaml_get_long(server_info, "uptime", 0));

	chat_interface_say(server, "*** %s/%s up for %s\n",
			   server_prog, version, uptime);
	if (text && strlen(text))
		chat_interface_say(server, "*** %s\n", text);

	g_free(text);
	g_free(uptime);
}

/* do_motd
 * Handle motd command
 */
static void
do_motd(Say2Server *server, const EnamlBlock *motd)
{
	gchar *text;

	g_assert(server);

	if (!motd)
		return;

	text = utf8_decode(enaml_get_string(motd, "text", NULL));
	if (text)
		chat_interface_say(server, "*** [MOTD] %s\n", text);

	g_free(text);
}

/* do_welcome
 * Handle welcome command
 */
static void
do_welcome(Say2Server *server, const EnamlBlock *welcome)
{
	const gchar *nick, *id;

	g_assert(server);

	if (!welcome) {
		chat_interface_say(server, "*** Welcome!\n");
		return;
	}

	nick = enaml_get_string(welcome, "nick", server->nickname);
	id = enaml_get_string(welcome, "id", server->id);

	if (nick && (!server->nickname || strcmp(server->nickname, nick))) {
		g_free(server->nickname);
		server->nickname = g_strdup(nick);
	}

	if (id && (!server->id || strcmp(server->id, id))) {
		g_free(server->id);
		server->id = g_strdup(id);
	}

	chat_interface_say(server, "*** Welcome, %s !\n", nick);
	chat_interface_update_by_server(server);

	/* Ensure notify is on, and get list of clients (will inherently
	 * whois ourself)
	 */
	send_enaml_buildf(server->sock, "prefs", "notify:%s", "on");
	send_enaml_buildf(server->sock, "list-clients", "type:%s by-id",
			  "user");
	server->squelch_prefs++;
	server->squelch_members++;
	server->members_do_whois++;
}

/* do_bad_nick
 * Handle bad-nick command
 */
static void
do_bad_nick(Say2Server *server, const EnamlBlock *bad_nick)
{
	const gchar *nick;
	gchar       *text;

	g_assert(server);

	if (!bad_nick) {
		chat_interface_say(server, "*** Can't use nickname\n");
		return;
	}

	nick = enaml_get_string(bad_nick, "nick", server->nickname);
	text = utf8_decode(enaml_get_string(bad_nick, "text",
				            "(no reason given)"));

	chat_interface_say(server, "*** Can't use nickname (%s): %s\n",
			   nick, text);

	g_free(text);
}

/* do_ping
 * Handle ping command
 */
static void
do_ping(Say2Server *server, const EnamlBlock *ping)
{
	EnamlBlock *block;
	
	g_assert(server);

	/* I think there's more to be gained by echoing back the whole
	 * block than just replying with a blank one
	 */
	block = enaml_new();
	if (ping) {
		enaml_add_block(block, "ping",
				enaml_copy_block(ping), ENAML_FREE);
	} else {
		enaml_add_flag(block, "ping");
	}
	send_enaml_block(server->sock, block);
	enaml_free(block);
}

/* do_signon
 * Handle the signon command
 */
static void
do_signon(Say2Server *server, const EnamlBlock *signon)
{
	const gchar *nick, *id, *email;
	gchar        timestr[12];
	time_t       msgtime;

	g_assert(server);

	if (!signon) {
		chat_interface_say(server, "*** Bogus signon received\n");
		return;
	}

	id = enaml_get_string(signon, "id", "?");
	nick = enaml_get_string(signon, "nick", "?");
	email = enaml_get_string(signon, "email", "?");
	msgtime = (time_t)enaml_get_long(signon, "time", time(NULL));

	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	g_print("FIXME New User %s nick:%s\n", id, nick);

	chat_interface_say(server, "%s*** Signon: %s <%s>\n",
			   timestr, nick, email);

	/* Send a whois to get more information about them */
	send_enaml_buildf(server->sock, "whois", "id:%s type:%s icon",
			  id, "user");
	server->squelch_whois++;
}

/* do_signoff
 * Handle the signoff command
 */
static void
do_signoff(Say2Server *server, const EnamlBlock *signoff)
{
	const gchar *nick, *id;
	gchar        timestr[12];
	time_t       msgtime;

	g_assert(server);

	if (!signoff) {
		chat_interface_say(server, "*** Bogus signoff received\n");
		return;
	}

	id = enaml_get_string(signoff, "id", "?");
	nick = enaml_get_string(signoff, "nick", "?");
	msgtime = (time_t)enaml_get_long(signoff, "time", time(NULL));

	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	g_print("FIXME Remove User %s nick:%s\n", id, nick);

	chat_interface_say(server, "%s*** Signoff: %s\n", timestr, nick);
}

/* do_prefs
 * Handle the prefs command
 */
static void
do_prefs(Say2Server *server, const EnamlBlock *prefs)
{
	const gchar *notify;
	gchar       *tagline;
	gint         password;
	
	g_assert(server);

	if (!prefs) {
		chat_interface_say(server, "*** Bogus prefs received\n");
		return;
	}

	notify = enaml_get_string(prefs, "notify", "?");
	tagline = utf8_decode(enaml_get_string(prefs, "tagline", NULL));
	password = (enaml_type(prefs, "password-changed") != ENAML_KEY_UNDEF
		    ? 1 : 0);

	/* Save the tagline */
	if (tagline) {
		g_free(server->tagline);
		server->tagline = g_strdup(tagline);
	}
	
	if (!server->squelch_prefs) {
		/* Display new preferences */
		chat_interface_say(server, "*** Preferences: Notify = %s\n",
				   notify);
		if (tagline) {
			chat_interface_say(server, "*** Preferences: "
					   "Tagline = %s\n", tagline);
		}
		if (password)
			chat_interface_say(server, "*** Preferences: %s\n",
					   "Password Changed");
	} else {
		/* Don't display anything */
		server->squelch_prefs--;
	}

	g_free(tagline);
}

/* do_status
 * Handle the status command
 */
static void
do_status(Say2Server *server, const EnamlBlock *status)
{
	const gchar *id, *nick, *mode;
	gchar       *text, *who;

	g_assert(server);

	if (!status) {
		chat_interface_say(server, "*** Bogus status received\n");
		return;
	}

	id = enaml_get_string(status, "id", server->id);
	nick = enaml_get_string(status, "nick", server->nickname);
	mode = enaml_get_string(status, "mode", "active");
	text = utf8_decode(enaml_get_string(status, "text", NULL));

	if (!g_strcasecmp(id, server->id)) {
		/* Notification of our status changing */
		if (!g_strcasecmp(mode, "active")) {
			server->status = SAY2_STATUS_ACTIVE;
		} else if (!g_strcasecmp(mode, "busy")) {
			server->status = SAY2_STATUS_BUSY;
		} else if (!g_strcasecmp(mode, "away")) {
			server->status = SAY2_STATUS_AWAY;
		}
	
		/* Update our msg */
		g_free(server->status_msg);
		server->status_msg = g_strdup(text);

		who = g_strdup_printf("You are");
	} else {
		/* Someone elses status changing */
		who = g_strdup_printf("%s is", nick);
	}

	g_print("FIXME Update User Status %s nick:%s\n", id, nick);

	/* Better English Campaign approved <g> */
	if (!g_strcasecmp(mode, "active"))
		mode = "here";


	if (text) {
		chat_interface_say(server, "*** %s now %s: %s\n",
				   who, mode, text);
	} else {
		chat_interface_say(server, "*** %s now %s\n", who, mode);
	}
	
	g_free(who);
	g_free(text);
}

/* do_change_nick
 * Handle the change-nick command
 */
static void
do_change_nick(Say2Server *server, const EnamlBlock *change_nick)
{
	const gchar *id, *old_nick, *nick;
	gchar       *who;

	g_assert(server);

	if (!change_nick) {
		chat_interface_say(server, "*** Bogus change-nick received\n");
		return;
	}

	id = enaml_get_string(change_nick, "id", server->id);
	old_nick = enaml_get_string(change_nick, "old-nick", server->nickname);
	nick = enaml_get_string(change_nick, "nick", "?");

	if (!g_strcasecmp(id, server->id)) {
		/* Notification that our nickname changed ok */
		g_free(server->nickname);
		server->nickname = g_strdup(nick);

		who = g_strdup_printf("You are");
	} else {
		/* Someone else changing their nick */
		who = g_strdup_printf("%s is", old_nick);
	}

	g_print("FIXME Update User Nickname %s old-nick: %s nick:%s\n",
		id, old_nick, nick);

	chat_interface_say(server, "*** %s now known as %s\n", who, nick);

	g_free(who);
}

/* do_members
 * Handle the members command
 */
static void
do_members(Say2Server *server, const EnamlBlock *members)
{
	const gchar *type;
	EnamlList   *list;
	GString     *buffer;
	gint         do_whois;

	g_assert(server);

	if (!members) {
		chat_interface_say(server, "*** Bogus members received\n");
		return;
	}

	type = enaml_get_string(members, "type", "user");
	list = enaml_get_list(members, "list");

	/* If this is a user members list, and members_do_whois is set, this
	 * is most likely the reply to the list-clients done when we joined.
	 * In that case whois each user that we find
	 */
	if (server->members_do_whois && !g_strcasecmp(type, "user")) {
		do_whois = 1;
		server->members_do_whois--;
	} else {
		do_whois = 0;
	}

	buffer = g_string_new("");

	if (list) {
		gint i, count;

		count = enaml_list_length(list);
		for (i = 0; i < count; i++) {
			const gchar *member;

			member = enaml_list_get_string(list, i, "?");
			g_string_sprintfa(buffer, " %s", member);

			/* Whois each member, we know that this means that
			 * members is a list of ids rather than nicks
			 */
			if (do_whois) {
				g_print("FIXME Ensure User %s\n", member);
				send_enaml_buildf(server->sock, "whois",
						  "id:%s type:%s icon",
						  member, "user");
				server->squelch_whois++;
			}
		}
	}

	if (!server->squelch_members) {
		/* Display the members */
		chat_interface_say(server, "*** Members:%s\n", buffer->str);
	} else {
		server->squelch_members--;
	}

	g_string_free(buffer, TRUE);
}

/* do_begin_response
 * Handle the begin-response command
 */
static void
do_begin_response(Say2Server *server, const EnamlBlock *begin_response)
{
	const gchar *response;
	glong        count, total;
	
	g_assert(server);

	if (!begin_response) {
		chat_interface_say(server, "*** Bogus "
				           "begin-response received\n");
		return;
	}

	response = enaml_get_string(begin_response, "response", "?");
	count = enaml_get_long(begin_response, "count", 0);
	total = enaml_get_long(begin_response, "total", 0);

	if (!g_strcasecmp(response, "client-info")) {
		/* Whom reply */

		/* Ditch any current one */
		whois_info_free_list(server->whom_list);

		/* Fill in info */
		server->in_whom = TRUE;
		server->whom_list = NULL;
		server->whom_count = count;
		server->whom_total = total;

	} else if (!g_strcasecmp(response, "say")) {
		/* History reply */
		server->in_history = TRUE;
		chat_interface_say(server, "*** History begins:\n");

	} else {
		/* Sodomy non sapiens */
		chat_interface_say(server, "*** Unknown response block (%s)\n",
				   response);
	}
}

/* do_end_response
 * Handle the end-response command
 */
static void
do_end_response(Say2Server *server, const EnamlBlock *end_response)
{
	const gchar *response;
	
	g_assert(server);

	if (!end_response) {
		chat_interface_say(server, "*** Bogus end-response received\n");
		return;
	}

	response = enaml_get_string(end_response, "response", "?");

	if (!g_strcasecmp(response, "client-info")) {
		/* End of Whom */
		whois_info_dump_list(server, server->whom_list,
				     server->whom_count, server->whom_total);

		/* Clean up */
		whois_info_free_list(server->whom_list);
		server->whom_list = NULL;
		server->in_whom = FALSE;

	} else if (!g_strcasecmp(response, "say")) {
		/* End of History */
		server->in_history = FALSE;
		chat_interface_say(server, "*** End of History.\n");

	} else {
		/* Sodomy non sapiens */
		chat_interface_say(server, "*** Unknown response "
				           "block end (%s)\n", response);
	}
}

/* do_client_info
 * Handle the client-info command
 */
static void
do_client_info(Say2Server *server, const EnamlBlock *client_info)
{
	WhoisInfo *wi;

	g_assert(server);

	if (!client_info) {
		chat_interface_say(server, "*** Bogus client-info received\n");
		return;
	}

	/* Reap the information */
	wi = whois_info_new_from_block(client_info);

	/* If it's a user, we want to update their information with the new
	 * stuff
	 */
	if (!g_strcasecmp(wi->type, "user")) {
		g_print("FIXME Update User %s\n", wi->id);

		/* User whois's may need to be squelched */
		if (server->squelch_whois) {
			server->squelch_whois--;
			goto out;
		}
	}

	if (server->in_whom) {
		/* Actually part of a WHOM */
		server->whom_list = g_slist_append(server->whom_list, wi);
		return;
	} else {
		/* Display it */
		whois_info_dump(server, wi);
	}

out:
	whois_info_free(wi);
}

/* do_not_present
 * Handle the not-present command
 */
static void
do_not_present(Say2Server *server, const EnamlBlock *not_present)
{
	const gchar *id, *nick, *type;

	g_assert(server);

	if (!not_present) {
		chat_interface_say(server, "*** Not logged in.\n");
		return;
	}

	id = enaml_get_string(not_present, "id", "?");
	nick = enaml_get_string(not_present, "nick", NULL);
	type = enaml_get_string(not_present, "type", "user");

	chat_interface_say(server, "*** %s is not logged in.\n",
			   (nick ? nick : id));
}

/* do_no_such_client
 * Handle the no-such-client command
 */
static void
do_no_such_client(Say2Server *server, const EnamlBlock *no_such_client)
{
	const gchar *id, *type;

	g_assert(server);

	if (!no_such_client) {
		chat_interface_say(server, "*** No such client.\n");
		return;
	}

	id = enaml_get_string(no_such_client, "id", "?");
	type = enaml_get_string(no_such_client, "type", "user");

	chat_interface_say(server, "*** No such %s: %s\n", type, id);
}

#if 0
/* do_client_attributes
 * Handle the client-attributes command
 */
static void
do_client_attributes(Say2Server *server, const EnamlBlock *client_attributes)
{
	/* FIXME */
}

/* do_set_attributes
 * Handle the set-attributes command
 */
static void
do_set_attributes(Say2Server *server, const EnamlBlock *set_attributes)
{
	/* FIXME */
}
#endif

/* do_say
 * Handle the say command
 */
static void
do_say(Say2Server *server, const EnamlBlock *say)
{
	const gchar *id, *nick;
	gchar       *text, *act;
	gchar        timestr[12];
	time_t       msgtime;

	g_assert(server);

	if (!say) {
		chat_interface_say(server, "*** Bogus say received\n");
		return;
	}

	id = enaml_get_string(say, "id", "?");
	nick = enaml_get_string(say, "nick", "?");
	text = utf8_decode(enaml_get_string(say, "text", NULL));
	act = utf8_decode(enaml_get_string(say, "act", NULL));
	msgtime = (time_t)enaml_get_long(say, "time", time(NULL));

	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	if (act)
		chat_interface_say(server, "%s* %s %s\n", timestr, nick, act);
	if (text)
		chat_interface_say(server, "%s<%s> %s\n", timestr, nick, text);

	g_free(text);
	g_free(act);
}

/* do_message
 * Handle the message command
 */
static void
do_message(Say2Server *server, const EnamlBlock *say)
{
	const gchar *id, *nick, *to_nick, *to_id;
	gchar       *text, *act;
	gchar        timestr[12];
	time_t       msgtime;

	g_assert(server);

	if (!say) {
		chat_interface_say(server, "*** Bogus say received\n");
		return;
	}

	id = enaml_get_string(say, "id", "?");
	nick = enaml_get_string(say, "nick", "?");
	to_id = enaml_get_string(say, "to-id", "?");
	to_nick = enaml_get_string(say, "to-nick", "?");
	text = utf8_decode(enaml_get_string(say, "text", NULL));
	act = utf8_decode(enaml_get_string(say, "act", NULL));
	msgtime = (time_t)enaml_get_long(say, "time", time(NULL));

	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	/* From me */
	if (!g_strcasecmp(id, server->id)) {
		if (act) {
			chat_interface_say(server, "%s-> [%s] * %s %s\n",
					   timestr, to_nick, nick, act);
		}
		if (text) {
			chat_interface_say(server, "%s-> [%s] %s\n",
					   timestr, to_nick, text);
		}
	}

	/* To me */
	if (!g_strcasecmp(to_id, server->id)) {
		if (act) {
			chat_interface_say(server, "%s* [%s] %s\n",
					   timestr, nick, act);
		}
		if (text) {
			chat_interface_say(server, "%s[%s] %s\n",
					   timestr, nick, text);
		}
	}

	g_free(text);
	g_free(act);
}

/* do_wall
 * Handle the wall command
 */
static void
do_wall(Say2Server *server, const EnamlBlock *wall)
{
	const gchar *id, *nick;
	gchar       *text;
	gchar        timestr[12];
	time_t       msgtime;

	g_assert(server);

	if (!wall) {
		chat_interface_say(server, "*** Bogus wall received\n");
		return;
	}

	id = enaml_get_string(wall, "id", "?");
	nick = enaml_get_string(wall, "nick", "?");
	text = utf8_decode(enaml_get_string(wall, "text", NULL));
	msgtime = (time_t)enaml_get_long(wall, "time", time(NULL));

	strcpy(timestr, "");
	strftime(timestr, sizeof timestr, "[%H:%M:%S] ", localtime(&msgtime));

	chat_interface_say(server, "%s*** WALL from %s: %s\n",
			   timestr, nick, text);

	g_free(text);
}


/* say2_server_connect
 * Connect to a say2 server described by the values returned from a connection
 * dialog.  Most of this is just copying things around in memory.
 */
void
say2_server_connect(ConnectValues *values, Say2Server *server)
{
	const struct servent *servent;

	g_assert(values);
	g_assert(server);

	/* Get the numerical value of the port */
	servent = getservbyname(values->server_port, "tcp");
	if (servent) {
		server->server_port = servent->s_port;
	} else {
		server->server_port = atoi(values->server_port);
	}

	/* Copy in the other string values */
	server->server_name = g_strdup(values->server_name);
	server->nickname = g_strdup(values->nickname);
	if (values->as_guest) {
		server->email = g_strdup(values->email);
		if (strlen(values->tagline))
			server->tagline = g_strdup(values->tagline);
	} else {
		server->password = g_strdup(values->password);
	}

	/* Time to connect */
	chat_interface_say(server, "*** Connecting to %s port %d...\n",
			   server->server_name, server->server_port);

	/* Create the socket and connect it */
	server->sock = socket_new(NULL, 0);
	if (!server->sock) {
		chat_interface_say(server, "*** Failed to make socket. :(\n");
		goto error;
	}
	socket_set_data(server->sock, server);
	if (!socket_connect(server->sock, server->server_name,
			    server->server_port, &say2_server_ready)) {
		chat_interface_say(server, "*** Failed to connect. :(\n");
		goto error;
	}

	/* Go away */
	return;

error:
	say2_server_free_values(server);
}

/* say2_server_read
 * Called when there's data on the socket.  Reads it, parses it and dispatches
 * it to the right function to handle the command.
 */
static void
say2_server_read(Socket *sock)
{
	static void *dispatch_table = NULL;
	Say2Server *server;

	/* Oh god, I'm turning into poskanser */
	if (!dispatch_table)
		dispatch_table = dispatch_create(command_table);
	
	/* Get the server structure off the socket */
	server = socket_get_data(sock);
	g_assert(server);
	g_assert(server->sock == sock);

	/* Loop through any lines available */
	for (;;) {
		EnamlBlock  *block;
		EnamlIter    iter;
		gchar       *line;
		const gchar *key;

		/* Get a line from the socket. */
		line = socket_getline(sock);
		if (!line)
			break;

		/* Parse the enaml in the line */
		block = enaml_parse(line);
		if (!block) {
			/* FIXME: do something intelligent */
			chat_interface_say(server, "*** Bad Enaml: %s\n", line);
			goto out;
		}

		/* A line can contain lots of different enaml keys and blocks.
		 * Iterate them so we don't miss one
		 */
		for (key = enaml_iter_first(block, &iter); key;
		     key = enaml_iter_next(&iter)) {
			const EnamlBlock *sub;

			/* Get the block and try to dispatch it */
			sub = enaml_get_block(block, key);
			if (!dispatch_go(dispatch_table, key,
					 server, (void *)sub, NULL)) {
				/* FIXME: do something intelligent */
				if (sub) {
					gchar *s;

					s = enaml_compose(block, FALSE);
					chat_interface_say(server, "*** Unknown"
							   " Command %s: %s\n",
							   key, s);
					g_free(s);
				} else {
					chat_interface_say(server, "*** Unknown"
							   " Command %s\n",
							   key);
				}

				goto out;
			}
		}

out:
		/* Free up what we did */
		enaml_free(block);
		g_free(line);
	}
}

/* say2_server_ready
 * Called when a connection to the server finishes (either successfully or
 * not)
 */
static void
say2_server_ready(Socket *sock, gint success)
{
	Say2Server *server;

	/* Get the server structure off the socket */
	server = socket_get_data(sock);
	g_assert(server);
	g_assert(server->sock == sock);

	/* Connection failed */
	if (!success) {
		chat_interface_say(server, "*** Failed to connect. :(\n");
		socket_close(sock);
		server->sock = 0;
		return;
	}

	chat_interface_say(server, "*** Connected!\n");

	socket_set_eof_fn(sock, &say2_server_eof);
	socket_set_read_fn(sock, &say2_server_read);
}

/* say2_server_eof
 * Called when the server drops our connection.
 */
static void
say2_server_eof(Socket *sock)
{
	Say2Server *server;

	/* Get the server structure off the socket */
	server = socket_get_data(sock);
	g_assert(server);
	g_assert(server->sock == sock);

	chat_interface_say(server, "*** Server dropped link.\n");

	server->sock = 0;
}

/* say2_server_new
 * Create a Say2Server structure
 */
Say2Server *
say2_server_new(void)
{
	return g_new0(Say2Server, 1);
}

/* say2_server_free_values
 * Free the values inside a Say2Server structure
 */
static void
say2_server_free_values(Say2Server *server)
{
	if (server->sock) {
		socket_close(server->sock);
		server->sock = 0;
	}

	g_free(server->server_name);
	g_free(server->nickname);
	g_free(server->password);
	g_free(server->email);
	g_free(server->tagline);

	server->server_name = server->nickname = server->password = 0;
	server->email = server->tagline = 0;
}
/* say2_server_free
 * Free a Say2Server structure
 */
void
say2_server_free(Say2Server *server)
{
	say2_server_free_values(server);
	g_free(server);
}
