/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * unicode.c
 *  - Conversion to and from UTF-8
 * -- 
 * $Id: unicode.c,v 1.5 2001/06/07 12:33:53 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <errno.h>

#include <glib.h>

#include <meko.h>

#ifdef HAVE_ICONV_H
#  include <iconv.h>
#endif /* HAVE_ICONV_H */

#include "unicode.h"

/* Forward prototypes */
static gchar *utf8_recode(iconv_t, const gchar *, double);


/* utf8_recode
 * Convert a string from one encoding to another dynamically adding buffer
 * space until we can fit the whole recoded string into it
 */
static gchar *
utf8_recode(iconv_t cd, const gchar *in, double mult)
{
	size_t outsz, inc;

#ifdef HAVE_ICONV
	if (cd == (iconv_t)-1)
		return g_strdup(in);

	inc = (int)(mult * strlen(in));
	outsz = inc + 1;

	/* Keep working until we succeed */
	for (;;) {
		gchar  *out, *outbuf, *inbuf;
		size_t  outlen, inlen;
		gint    ret;

		/* Reset iconv */
		iconv(cd, NULL, NULL, NULL, NULL);

		/* Prepare buffers */
		outbuf = out = g_new0(gchar, outsz);
		inbuf = (gchar *)in;
		inlen = strlen(in);
		outlen = outsz;

		/* Try to convert */
		ret = iconv(cd, &inbuf, &inlen, &outbuf, &outlen);
		if ((ret >= 0) || (errno != E2BIG))
			return out;

		/* Not enough room, increase buffer and go round again */
		g_free(out);
		outsz += inc;
	}
#else /* HAVE_ICONV */
	return NULL;
#endif /* HAVE_ICONV */
}

/* utf8_encode
 * Convert a string from ISO-8859-1 to UTF-8 using utf8_recode
 */
gchar *
utf8_encode(const gchar *str)
{
	gchar   *ret;
	iconv_t  cd;

	if (!str)
		return NULL;

#ifdef HAVE_ICONV
	cd = iconv_open(ENCODING_UTF8, ENCODING_DEFAULT);
	ret = utf8_recode(cd, str, 1.2);
	iconv_close(cd);

	return ret;
#else /* HAVE_ICONV */
	return g_strdup(str);
#endif /* HAVE_ICONV */
}

/* utf8_decode
 * Convert a string from UTF-8 to ISO-8859-1 using utf8_recode
 */
gchar *
utf8_decode(const gchar *str)
{
	gchar   *ret;
	iconv_t  cd;

	if (!str)
		return NULL;

#ifdef HAVE_ICONV
	cd = iconv_open(ENCODING_DEFAULT, ENCODING_UTF8);
	ret = utf8_recode(cd, str, 1);
	iconv_close(cd);

	return ret;
#else /* HAVE_ICONV */
	return g_strdup(str);
#endif /* HAVE_ICONV */
}
