/* meko
 * Copyright (C) 2001 Scott James Remnant <scott@netsplit.com>.
 * All Rights Reserved.
 *
 * whois_info.c
 *  - Handles the large amount of information needed for a whois or whom
 *    reply
 * -- 
 * $Id: whois_info.c,v 1.1 2001/06/07 12:33:30 scott Exp $
 * 
 * This file is distributed according to the GNU General Public
 * License.  For full details, read the top of 'main.c' or the
 * file called COPYING that was distributed with this code.
 */

#include <time.h>

#include <glib.h>
#include <say2/say2.h>

#include <meko.h>

#include "whois_info.h"
#include "say2_server.h"
#include "chat_interface.h"
#include "unicode.h"

/* whois_info_dump
 * Display the contents of a WhoisInfo struct to a ChatInterface in the
 * style of a /whois request
 */
void
whois_info_dump(Say2Server *server, WhoisInfo *wi)
{
	GString     *buffer;

	g_assert(server);
	g_assert(wi);

	/* The first line needs some building up */
	buffer = g_string_new("");
	g_string_sprintfa(buffer, "%s <%s>", (wi->nick ? wi->nick : wi->id),
			  wi->email);
	if (wi->nick && g_strcasecmp(wi->id, wi->nick))
		g_string_sprintfa(buffer, " (normally: %s)", wi->id);
	if (wi->offline)
		g_string_sprintfa(buffer, " [OFFLINE]");
	if (wi->mode && !g_strcasecmp(wi->mode, "idle"))
		g_string_sprintfa(buffer, " [IDLE]");
	chat_interface_say(server, "*** %s\n", buffer->str);
	g_string_free(buffer, TRUE);

	if (wi->host) {
		chat_interface_say(server, "***     Connected from %s\n",
				   wi->host);
	}

	if (wi->tagline) {
		chat_interface_say(server, "***     Tagline: %s\n",
				   wi->tagline);
	}

	if (wi->client) {
		chat_interface_say(server, "***     Client: %s\n",
				   wi->client);
	}
	
	if (wi->login) {
		gchar timestr[40];

		strftime(timestr, sizeof timestr, "%d-%b-%Y %H:%M:%S",
			 localtime(&(wi->login)));
		chat_interface_say(server, "***     Logged in at %s\n",
				   timestr);
	}

	if (wi->idle && (wi->idle > 180)) {
		gchar *idlestr;

		idlestr = whois_timespan_str(wi->idle);
		chat_interface_say(server, "***     Idle for %s\n", idlestr);
		g_free(idlestr);
	}

	/* If they are AWAY or BUSY, say so.  Give their status message if
	 * they've got one as well.
	 */
	if (wi->mode) {
		const gchar *modestr;

		if (!g_strcasecmp(wi->mode, "away")) {
			modestr = "AWAY";
		} else if (!g_strcasecmp(wi->mode, "busy")) {
			modestr = "BUSY";
		} else {
			modestr = 0;
		}

		if (modestr) {
			chat_interface_say(server, "***     %s%s%s\n", modestr,
					   (wi->status ? ": " : ""),
					   (wi->status ? wi->status : ""));
		} else if (wi->status) {
			chat_interface_say(server, "***     %s is %s\n",
					   (wi->nick ? wi->nick : wi->id),
					   wi->status);
		}
	}

	/* If last_logout = 0 it means that it was present and set to 0, the
	 * default is -1.  This means they've never logged in.
	 */
	if (wi->last_logout > 0) {
		gchar   timestr[40];
		gchar  *agostr;
		time_t  now;

		now = time(NULL);
		agostr = whois_timespan_str(now - wi->last_logout);
		
		strftime(timestr, sizeof timestr, "%d-%b-%Y %H:%M:%S",
			 localtime(&(wi->last_logout)));

		chat_interface_say(server, "***     Last logged out "
				           "%s (%s ago)\n", timestr, agostr);

		g_free(agostr);

	} else if (wi->last_logout == 0) {
		chat_interface_say(server, "***     Never logged in.\n");
	}
}

/* whois_info_dump_list
 * Display the contents of a list of WhoisInfo structs to a ChatInterface
 * in the style of a /whom request
 */
void
whois_info_dump_list(Say2Server *server, GSList *list, glong count, glong total)
{
	gint     max_nick, max_email;
	GSList  *iter;
	
	g_assert(server);

	if (!count || !total || !list) {
		chat_interface_say(server, "*** 0 users logged in. (?!)\n");
		return;
	}

	/* Not all were returned */
	if (count < total) {
		chat_interface_say(server,
				   "*** %ld user%s online (%ld shown):\n",
				   total, (total == 1 ? "" : "s"), count);
	} else {
		chat_interface_say(server, "*** %ld user%s online:\n",
				   count, (count == 1 ? "" : "s"));
	}

	/* Figure out how spaced out we need to make things */
	max_nick = max_email = 0;
	for (iter = list; iter; iter = g_slist_next(iter)) {
		WhoisInfo *wi;

		wi = (WhoisInfo *)iter->data;
		g_assert(wi);

		if (wi->nick) {
			if (strlen(wi->nick) > max_nick)
				max_nick = strlen(wi->nick);
		} else {
			if (strlen(wi->id) > max_nick)
				max_nick = strlen(wi->id);
		}

		if (strlen(wi->email) > max_email)
			max_email = strlen(wi->email);
	}
	
	/* Output each entry */
	for (iter = list; iter; iter = g_slist_next(iter)) {
		GString   *buffer;
		WhoisInfo *wi;
		
		wi = (WhoisInfo *)iter->data;
		g_assert(wi);

		/* Build up the first line */
		buffer = g_string_new("");
		g_string_sprintfa(buffer, "%-*s <%s>%*s",
				  max_nick, wi->nick, wi->email,
				  (max_email - strlen(wi->email)), "");
		if (wi->login) {
			gchar  timestr[8];
			time_t now;

			now = time(NULL);
			if ((now - wi->login) > 43200) {
				strftime(timestr, sizeof timestr, "%d-%b",
					 localtime(&(wi->login)));
			} else {
				strftime(timestr, sizeof timestr, "%H:%M",
					 localtime(&(wi->login)));
			}
			g_string_sprintfa(buffer, "  login @ %-7s", timestr);
		}
		if (!g_strcasecmp(wi->mode, "idle")) {

			g_string_sprintfa(buffer, " [idle");

			if (wi->idle) {
				gchar *spanstr;

				spanstr = whois_timespan_str(wi->idle);
				g_string_sprintfa(buffer, " %s", spanstr);
				g_free(spanstr);
			}

			g_string_append(buffer, "]");

		} else if (!wi->status
			   && (!g_strcasecmp(wi->mode, "away")
			       || !g_strcasecmp(wi->mode, "busy"))) {

			if (!g_strcasecmp(wi->mode, "away")) {
				g_string_sprintfa(buffer, " [AWAY");
			} else {
				g_string_sprintfa(buffer, " [BUSY");
			}

			if (wi->status_time) {
				gchar *spanstr;

				spanstr = whois_timespan_str(wi->status_time);
				g_string_sprintfa(buffer, " %s", spanstr);
				g_free(spanstr);
			}

			g_string_append(buffer, "]");
		}
		chat_interface_say(server, "***     %s\n", buffer->str);
		g_string_free(buffer, TRUE);

		/* Add the status text if there is some */
		if (wi->status) {
			const gchar *modestr;

			if (!g_strcasecmp(wi->mode, "active")) {
				modestr = "here";
			} else {
				modestr = wi->mode;
			}

			chat_interface_say(server, "***         %s: %s\n",
					   modestr, wi->status);
		}
	}
}

/* whois_timespan_str
 * Returns an allocated string that texually says how much time is in its
 * arg
 */
gchar *
whois_timespan_str(time_t secs)
{
	time_t   days, hrs, mins;
	GString *buffer;
	gchar   *ret;

	buffer = g_string_new("");

	days = (time_t)(secs / 86400);
	secs %= 86400;
	if (days)
		g_string_sprintfa(buffer, "%ld day%s ", days,
				  (days == 1 ? "" : "s"));

	hrs = (time_t)(secs / 3600);
	secs %= 3600;
	if (hrs)
		g_string_sprintfa(buffer, "%ld hr ", hrs);

	mins = (time_t)(secs / 60);
	secs %= 60;
	if (mins && !days)
		g_string_sprintfa(buffer, "%ld min ", mins);

	/* Don't display anything smaller than a minute */
	if (!buffer->len)
		g_string_sprintfa(buffer, "<1 min ");

	/* Removes the trailing space */
	ret = g_strndup(buffer->str, buffer->len - 1);
	g_string_free(buffer, TRUE);

	return ret;
}

/* whois_info_new_from_block
 * Allocate room for a new WhoisInfo struct filling it with values taken
 * from an EnamlBlock
 */
WhoisInfo *
whois_info_new_from_block(const EnamlBlock *client_info)
{
	WhoisInfo *wi;

	wi = whois_info_new();

	wi->type = g_strdup(enaml_get_string(client_info, "type", "user"));

	wi->id = g_strdup(enaml_get_string(client_info, "id", "?"));
	wi->nick = g_strdup(enaml_get_string(client_info, "nick", NULL));
	wi->email = g_strdup(enaml_get_string(client_info, "email", "?"));
	wi->host = g_strdup(enaml_get_string(client_info, "host", NULL));
	wi->address = g_strdup(enaml_get_string(client_info, "address", NULL));
	wi->tagline = utf8_decode(enaml_get_string(client_info,
				                   "tagline", NULL));
	wi->client = utf8_decode(enaml_get_string(client_info, "client", NULL));

	wi->offline = (enaml_type(client_info, "offline") != ENAML_KEY_UNDEF
		       ? TRUE : FALSE);

	wi->mode = g_strdup(enaml_get_string(client_info, "mode", NULL));
	wi->status = utf8_decode(enaml_get_string(client_info, "status", NULL));
	wi->status_time = (time_t)enaml_get_long(client_info, "status-time", 0);

	wi->login = (time_t)enaml_get_long(client_info, "login", 0);
	wi->idle = (time_t)enaml_get_long(client_info, "idle", 0);
	wi->last_login = (time_t)enaml_get_long(client_info, "last-login", -1);
	wi->last_logout = (time_t)enaml_get_long(client_info,
			                         "last-logout", -1);

	return wi;
}

/* whois_info_new
 * Allocate room for a new WhoisInfo struct
 */
WhoisInfo *
whois_info_new(void)
{
	return g_new0(WhoisInfo, 1);
}

/* whois_info_free
 * Free's up the space used by a WhoisInfo struct
 */
void
whois_info_free(WhoisInfo *wi)
{
	if (!wi)
		return;

	g_free(wi->type);
	g_free(wi->id);
	g_free(wi->nick);
	g_free(wi->email);
	g_free(wi->host);
	g_free(wi->address);
	g_free(wi->tagline);
	g_free(wi->client);
	g_free(wi->mode);
	g_free(wi->status);
	g_free(wi->icon);
	g_free(wi);
}

/* whois_info_free_list
 * Free up a list of WhoisInfo structs
 */
void
whois_info_free_list(GSList *list)
{
	GSList *iter;

	if (!list)
		return;

	for (iter = list; iter; iter = g_slist_next(iter)) {
		whois_info_free((WhoisInfo *)list->data);
		list->data = NULL;
	}

	g_slist_free(list);
}
